@@ -14,8 +14,7 @@
  *  GNU General Public License for more details.
  *
  *  You should have received a copy of the GNU General Public License
- *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 
 #include "isync.h"
@@ -28,7 +27,6 @@
 #include <stdio.h>
 #include <unistd.h>
 #include <sys/stat.h>
-#include <errno.h>
 #include <time.h>
 
 #include <db.h>
@@ -73,7 +71,7 @@ convert( config_t *box )
 	FILE *fp;
 	msg_t *msgs;
 	DB *db;
-	int i, ret, fd, uidval, maxuid, bl, uid, rmsgs, nmsgs, uv[2];
+	int i, ret, fd, uidval, maxuid, uid, rmsgs, nmsgs, uv[2];
 	unsigned u;
 	struct stat sb;
 	char buf[_POSIX_PATH_MAX], diumname[_POSIX_PATH_MAX],
@@ -97,10 +95,9 @@ convert( config_t *box )
 	for (i = 0; i < 3; i++) {
 		nfsnprintf( buf, sizeof(buf), "%s/%s", mboxdir, subdirs[i] );
 		if (stat( buf, &sb )) {
-			fprintf( stderr, "ERROR: stat %s: %s (errno %d)\n", buf,
-			         strerror(errno), errno );
+			sys_error( "ERROR: cannot access %s", buf );
 			fprintf( stderr,
-			         "ERROR: %s does not appear to be a valid maildir style mailbox\n",
+			         "ERROR: '%s' does not appear to be a valid maildir style mailbox\n",
 			         mboxdir );
 			goto err1;
 		}
@@ -111,7 +108,7 @@ convert( config_t *box )
 	nfsnprintf( sname, sizeof(sname), "%s/.mbsyncstate", mboxdir );
 
 	if ((fd = open( ilname, O_WRONLY|O_CREAT, 0600 )) < 0) {
-		perror( ilname );
+		sys_error( "Cannot create %s", ilname );
 		goto err1;
 	}
 #if SEEK_SET != 0
@@ -121,20 +118,20 @@ convert( config_t *box )
 	lck.l_type = F_WRLCK;
 #endif
 	if (fcntl( fd, F_SETLKW, &lck )) {
-		perror( ilname );
+		sys_error( "Cannot lock %s", ilname );
 	  err2:
 		close( fd );
 		goto err1;
 	}
 
 	if (!(fp = fopen( iuvname, "r" ))) {
-		perror( iuvname );
+		sys_error( "Cannot open %s", iuvname );
 		goto err2;
 	}
 	fscanf( fp, "%d", &uidval );
 	fclose( fp );
 	if (!(fp = fopen( imuname, "r" ))) {
-		perror( imuname );
+		sys_error( "Cannot open %s", imuname );
 		goto err2;
 	}
 	fscanf( fp, "%d", &maxuid );
@@ -160,12 +157,11 @@ convert( config_t *box )
 	rmsgs = 0;
 	nmsgs = 0;
 	for (i = 0; i < 2; i++) {
-		bl = nfsnprintf( buf, sizeof(buf), "%s/%s/", mboxdir, subdirs[i] );
+		nfsnprintf( buf, sizeof(buf), "%s/%s/", mboxdir, subdirs[i] );
 		if (!(d = opendir( buf ))) {
-			perror( "opendir" );
+			sys_error( "Cannot list %s", buf );
 		  err4:
-			if (msgs)
-				free( msgs );
+			free( msgs );
 			if (db)
 				db->close( db, 0 );
 			goto err2;
@@ -200,7 +196,7 @@ convert( config_t *box )
 	qsort( msgs, nmsgs, sizeof(msg_t), compare_uids );
 
 	if (!(fp = fopen( sname, "w" ))) {
-		perror( sname );
+		sys_error( "Cannot create %s", sname );
 		goto err4;
 	}
 	if (box->max_messages) {
@@ -239,7 +235,7 @@ convert( config_t *box )
 		rename( iumname, diumname );
 	} else {
 		if (!(fp = fopen( uvname, "w" ))) {
-			perror( uvname );
+			sys_error( "Cannot create %s", uvname );
 			goto err4;
 		}
 		fprintf( fp, "%d\n%d\n", uidval, maxuid );
@@ -252,8 +248,7 @@ convert( config_t *box )
 	close( fd );
 	unlink( ilname );
 
-	if (msgs)
-		free( msgs );
+	free( msgs );
 	free( mboxdir );
 	return;
 }
